# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from io import StringIO
from unittest.mock import patch

import numpy as np
import pytest
import tensorflow as tf
from packaging.version import parse
from tensorboard.plugins.hparams import api as hparams_api
from tensorflow import keras

import keras_tuner
from keras_tuner import errors
from keras_tuner.engine import tuner as tuner_module

INPUT_DIM = 2
NUM_CLASSES = 3
NUM_SAMPLES = 64
TRAIN_INPUTS = np.random.random(size=(NUM_SAMPLES, INPUT_DIM))
TRAIN_TARGETS = np.random.randint(0, NUM_CLASSES, size=(NUM_SAMPLES, 1))
VAL_INPUTS = np.random.random(size=(NUM_SAMPLES, INPUT_DIM))
VAL_TARGETS = np.random.randint(0, NUM_CLASSES, size=(NUM_SAMPLES, 1))


def build_model(hp):
    inputs = keras.Input(shape=(INPUT_DIM,))
    x = inputs
    for i in range(hp.Int("num_layers", 1, 4)):
        x = keras.layers.Dense(
            units=hp.Int(f"units_{str(i)}", 5, 9, 1, default=6),
            activation="relu",
        )(x)

    outputs = keras.layers.Dense(NUM_CLASSES, activation="softmax")(x)
    model = keras.Model(inputs, outputs)
    model.compile(
        optimizer=keras.optimizers.Adam(
            hp.Choice("learning_rate", [1e-2, 1e-3, 1e-4])
        ),
        loss="sparse_categorical_crossentropy",
        metrics=["accuracy"],
    )
    return model


class MockModel(keras.Model):
    def __init__(self, full_history):
        super().__init__()
        self.full_history = full_history
        self.callbacks = []
        self.optimizer = True

    def call_callbacks(self, callbacks, method_name, *args, **kwargs):
        for callback in callbacks:
            method = getattr(callback, method_name)
            method(*args, **kwargs)

    def on_epoch_begin(self, epoch):
        for callback in self.callbacks:
            callback.on_epoch_begin(epoch, logs=None)

    def on_epoch_end(self, epoch):
        logs = {"loss": np.average(self.full_history[epoch])}
        for callback in self.callbacks:
            callback.on_epoch_end(epoch, logs=logs)

    def on_batch_begin(self, epoch, batch):
        for callback in self.callbacks:
            callback.on_batch_begin(batch, logs=None)

    def on_batch_end(self, epoch, batch):
        logs = {"loss": self.full_history[epoch][batch]}
        for callback in self.callbacks:
            callback.on_batch_end(epoch, logs=logs)

    def fit(self, *args, **kwargs):
        self.callbacks = kwargs["callbacks"]
        for callback in self.callbacks:
            callback.model = self
        for epoch in range(len(self.full_history)):
            self.on_epoch_begin(epoch)
            for batch in range(len(self.full_history[epoch])):
                self.on_batch_begin(epoch, batch)
                self.on_batch_end(epoch, batch)
            self.on_epoch_end(epoch)
        history = keras.callbacks.History()
        history.history = {
            "loss": [
                np.average(epoch_values) for epoch_values in self.full_history
            ]
        }
        return history

    def save_weights(self, fname, **kwargs):
        pass

    def get_config(self):
        return {}


class MockHyperModel(keras_tuner.HyperModel):
    mode_0 = [[10, 9, 8], [7, 6, 5], [4, 3, 2]]
    mode_1 = [[13, 13, 13], [12, 12, 12], [11, 11, 11]]

    def __init__(self):
        # The first call to `build` in tuner __init__
        # will reset this to 0
        self.mode_0_execution_count = -1

    def build(self, hp):
        if hp.Choice("mode", [0, 1]) == 0:
            return MockModel(self.mode_0)
        return MockModel(self.mode_1)


def build_subclass_model(hp):
    class MyModel(keras.Model):
        def build(self, _):
            self.layer = keras.layers.Dense(NUM_CLASSES, activation="softmax")

        def call(self, x):
            x = x + hp.Float("bias", 0, 10)
            return self.layer(x)

        # Currently necessary, because we save the model.
        # Note that this model is not written w/ best practices,
        # because the hp.Float value of the best model cannot be
        # inferred from `get_config()`. The best practice is to pass
        # HPs as __init__ arguments to subclass Layers and Models.
        def get_config(self):
            return {}

    model = MyModel()
    model.compile(
        optimizer=keras.optimizers.Adam(
            hp.Choice("learning_rate", [1e-2, 1e-3, 1e-4])
        ),
        loss="sparse_categorical_crossentropy",
        metrics=["accuracy"],
    )
    return model


class ExampleHyperModel(keras_tuner.HyperModel):
    def build(self, hp):
        inputs = keras.Input(shape=(INPUT_DIM,))
        x = inputs
        for i in range(hp.Int("num_layers", 1, 4)):
            x = keras.layers.Dense(
                units=hp.Int(f"units_{str(i)}", 5, 9, 1, default=6),
                activation="relu",
            )(x)

        outputs = keras.layers.Dense(NUM_CLASSES, activation="softmax")(x)
        model = keras.Model(inputs, outputs)
        model.compile(
            optimizer=keras.optimizers.Adam(
                hp.Choice("learning_rate", [1e-2, 1e-3, 1e-4])
            ),
            loss="sparse_categorical_crossentropy",
            metrics=["accuracy"],
        )
        return model

    def fit(self, hp, model, *args, **kwargs):
        return model.fit(*args, shuffle=hp.Boolean("shuffle"), **kwargs)


def test_basic_tuner_attributes(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        build_model,
        objective="val_accuracy",
        max_trials=2,
        executions_per_trial=3,
        directory=tmp_path,
    )

    assert tuner.oracle.objective.name == "val_accuracy"
    assert tuner.oracle.max_trials == 2
    assert tuner.executions_per_trial == 3
    assert tuner.directory == tmp_path
    assert tuner.hypermodel.__class__.__name__ == "DefaultHyperModel"
    assert len(tuner.oracle.hyperparameters.space) == 3  # default search space
    assert len(tuner.oracle.hyperparameters.values) == 3  # default search space

    tuner.search_space_summary()

    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=2,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    tuner.results_summary()

    assert len(tuner.oracle.trials) == 2
    assert os.path.exists(os.path.join(str(tmp_path), "untitled_project"))


def test_multi_objective(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        build_model,
        objective=["val_accuracy", "val_loss"],
        max_trials=2,
        executions_per_trial=3,
        directory=tmp_path,
    )

    assert tuner.oracle.objective.name == "multi_objective"

    tuner.search_space_summary()

    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=2,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    tuner.results_summary()


def test_no_hypermodel_with_objective(tmp_path):
    class MyTuner(keras_tuner.tuners.RandomSearch):
        def run_trial(self, trial, *args, **kwargs):
            hp = trial.hyperparameters
            return {"val_loss": hp.Float("value", 0, 10)}

    tuner = MyTuner(
        objective="val_loss",
        max_trials=2,
        directory=tmp_path,
    )
    tuner.search()

    assert len(tuner.oracle.trials) == 2


def test_no_objective_with_hypermodel(tmp_path):
    class MyHyperModel(ExampleHyperModel):
        def fit(self, hp, model, *args, **kwargs):
            return hp.Float("value", 0, 10)

    tuner = keras_tuner.tuners.RandomSearch(
        hypermodel=MyHyperModel(),
        max_trials=2,
        executions_per_trial=3,
        directory=tmp_path,
    )
    tuner.search()

    assert len(tuner.oracle.trials) == 2


def test_no_hypermodel_no_objective(tmp_path):
    class MyTuner(keras_tuner.tuners.RandomSearch):
        def run_trial(self, trial, *args, **kwargs):
            hp = trial.hyperparameters
            return hp.Float("value", 0, 10)

    tuner = MyTuner(
        objective="val_loss",
        max_trials=2,
        directory=tmp_path,
    )
    tuner.search()

    assert len(tuner.oracle.trials) == 2


def test_no_hypermodel_without_override_run_trial_error(tmp_path):
    with pytest.raises(ValueError, match="Received `hypermodel=None`"):
        keras_tuner.tuners.RandomSearch(
            max_trials=2,
            executions_per_trial=3,
            directory=tmp_path,
        )


def test_fit_return_string(tmp_path):
    class MyHyperModel(ExampleHyperModel):
        def fit(self, hp, model, *args, **kwargs):
            return hp.Choice("value", ["a", "b"])

    tuner = keras_tuner.tuners.RandomSearch(
        objective="val_loss",
        hypermodel=MyHyperModel(),
        max_trials=2,
        executions_per_trial=3,
        directory=tmp_path,
    )

    with pytest.raises(TypeError, match="HyperModel\.fit\(\) to be one of"):
        tuner.search()


def test_run_trial_return_string(tmp_path):
    class MyTuner(keras_tuner.tuners.RandomSearch):
        def run_trial(self, trial, **kwargs):
            return trial.hyperparameters.Choice("value", ["a", "b"])

    tuner = MyTuner(
        objective="val_loss",
        max_trials=2,
        executions_per_trial=3,
        directory=tmp_path,
    )

    with pytest.raises(TypeError, match="Tuner\.run_trial\(\) to be one of"):
        tuner.search()


def test_no_objective_fit_return_not_float(tmp_path):
    class MyHyperModel(ExampleHyperModel):
        def fit(self, hp, model, *args, **kwargs):
            return {"val_loss": hp.Float("value", 0, 10)}

    tuner = keras_tuner.tuners.RandomSearch(
        hypermodel=MyHyperModel(),
        max_trials=2,
        executions_per_trial=3,
        directory=tmp_path,
    )

    with pytest.raises(
        TypeError, match="HyperModel\.fit\(\) to be a single float"
    ):
        tuner.search()


def test_no_objective_run_trial_return_not_float(tmp_path):
    class MyTuner(keras_tuner.tuners.RandomSearch):
        def run_trial(self, trial, **kwargs):
            return {"val_loss": trial.hyperparameters.Float("value", 0, 10)}

    tuner = MyTuner(
        max_trials=2,
        executions_per_trial=3,
        directory=tmp_path,
    )

    with pytest.raises(
        TypeError, match="Tuner\.run_trial\(\) to be a single float"
    ):
        tuner.search()


def test_callbacks_in_fit_kwargs(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        build_model,
        objective="val_accuracy",
        max_trials=2,
        executions_per_trial=3,
        directory=tmp_path,
    )
    with patch.object(
        tuner, "_build_and_fit_model", wraps=tuner._build_and_fit_model
    ) as mock_build_and_fit_model:
        tuner.search(
            x=TRAIN_INPUTS,
            y=TRAIN_TARGETS,
            epochs=2,
            validation_data=(VAL_INPUTS, VAL_TARGETS),
            callbacks=[
                keras.callbacks.EarlyStopping(),
                keras.callbacks.TensorBoard(tmp_path),
            ],
        )
        assert len(tuner.oracle.trials) == 2
        callback_class_names = [
            x.__class__.__name__
            for x in mock_build_and_fit_model.call_args[1]["callbacks"]
        ]
        assert callback_class_names == [
            "EarlyStopping",
            "TensorBoard",
            "Callback",
            "TunerCallback",
            "SaveBestEpoch",
        ]


def test_hypermodel_with_dynamic_space(tmp_path):
    hypermodel = ExampleHyperModel()
    tuner = keras_tuner.tuners.RandomSearch(
        hypermodel,
        objective="val_accuracy",
        max_trials=2,
        executions_per_trial=3,
        directory=tmp_path,
    )

    assert tuner.hypermodel == hypermodel

    tuner.search_space_summary()

    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=2,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    tuner.results_summary()

    assert len(tuner.oracle.trials) == 2
    tuner.oracle.hyperparameters.get("shuffle")


def test_override_compile(tmp_path):
    class MyHyperModel(ExampleHyperModel):
        def fit(self, hp, model, *args, **kwargs):
            history = super().fit(hp, model, *args, **kwargs)
            assert model.optimizer.__class__.__name__ == "RMSprop"
            assert model.loss == "mse"
            assert len(model.metrics) >= 2
            assert model.metrics[-2]._fn.__name__ == "mean_squared_error"
            assert (
                model.metrics[-1]._fn.__name__ == "sparse_categorical_accuracy"
            )
            return history

    tuner = keras_tuner.tuners.RandomSearch(
        MyHyperModel(),
        objective="val_mse",
        max_trials=2,
        executions_per_trial=1,
        metrics=["mse", "accuracy"],
        loss="mse",
        optimizer="rmsprop",
        directory=tmp_path,
    )

    assert tuner.oracle.objective.name == "val_mse"
    assert tuner.optimizer == "rmsprop"
    assert tuner.loss == "mse"
    assert tuner.metrics == ["mse", "accuracy"]

    tuner.search_space_summary()
    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=2,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )
    tuner.results_summary()
    model = tuner.get_best_models()[0]
    assert model.loss == "mse"


def test_override_optimizer_with_actual_optimizer_object(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        build_model,
        objective="val_loss",
        max_trials=4,
        optimizer=keras.optimizers.Adam(0.01),
        directory=tmp_path,
    )
    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=2,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )


def test_static_space(tmp_path):
    def build_model_static(hp):
        inputs = keras.Input(shape=(INPUT_DIM,))
        x = inputs
        for i in range(hp.get("num_layers")):
            x = keras.layers.Dense(
                units=hp.get(f"units_{str(i)}"), activation="relu"
            )(x)
        outputs = keras.layers.Dense(NUM_CLASSES, activation="softmax")(x)
        model = keras.Model(inputs, outputs)
        model.compile(
            optimizer=keras.optimizers.Adam(hp.get("learning_rate")),
            loss="sparse_categorical_crossentropy",
            metrics=["accuracy"],
        )
        return model

    hp = keras_tuner.HyperParameters()
    hp.Int("num_layers", 1, 3, 1, default=2)
    hp.Int("units_0", 4, 6, 1, default=5)
    hp.Int("units_1", 4, 6, 1, default=5)
    hp.Int("units_2", 4, 6, 1, default=5)
    hp.Choice("learning_rate", [0.01, 0.001])
    tuner = keras_tuner.tuners.RandomSearch(
        build_model_static,
        objective="val_accuracy",
        max_trials=4,
        directory=tmp_path,
        hyperparameters=hp,
        allow_new_entries=False,
    )

    assert tuner.oracle.hyperparameters == hp
    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=2,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )
    assert len(tuner.oracle.trials) == 4


def test_static_space_errors(tmp_path):
    def build_model_static(hp):
        inputs = keras.Input(shape=(INPUT_DIM,))
        x = inputs
        for i in range(hp.get("num_layers")):
            x = keras.layers.Dense(
                units=hp.get(f"units_{str(i)}"), activation="relu"
            )(x)
        outputs = keras.layers.Dense(NUM_CLASSES, activation="softmax")(x)
        model = keras.Model(inputs, outputs)
        model.compile(
            optimizer=keras.optimizers.Adam(
                hp.Float("learning_rate", 1e-5, 1e-2)
            ),
            loss="sparse_categorical_crossentropy",
            metrics=["accuracy"],
        )
        return model

    hp = keras_tuner.HyperParameters()
    hp.Int("num_layers", 1, 3, 1, default=2)
    hp.Int("units_0", 4, 6, 1, default=5)
    hp.Int("units_1", 4, 6, 1, default=5)

    with pytest.raises(RuntimeError, match="`allow_new_entries` is `False`"):
        tuner = keras_tuner.tuners.RandomSearch(
            build_model_static,
            objective="val_accuracy",
            max_trials=2,
            directory=tmp_path,
            hyperparameters=hp,
            allow_new_entries=False,
        )
        tuner.search(
            x=TRAIN_INPUTS,
            y=TRAIN_TARGETS,
            epochs=2,
            validation_data=(VAL_INPUTS, VAL_TARGETS),
        )

    def build_model_static_invalid(hp):
        inputs = keras.Input(shape=(INPUT_DIM,))
        x = inputs
        for i in range(hp.get("num_layers")):
            x = keras.layers.Dense(
                units=hp.get(f"units_{str(i)}"), activation="relu"
            )(x)
        outputs = keras.layers.Dense(NUM_CLASSES, activation="softmax")(x)
        model = keras.Model(inputs, outputs)
        model.compile(
            optimizer=keras.optimizers.Adam(
                hp.Float("learning_rate", 0.001, 0.008, 0.001)
            ),
            loss="sparse_categorical_crossentropy",
            metrics=["accuracy"],
        )
        return model

    with pytest.raises(RuntimeError, match="`allow_new_entries` is `False`"):
        tuner = keras_tuner.tuners.RandomSearch(
            build_model_static_invalid,
            objective="val_accuracy",
            max_trials=2,
            directory=tmp_path,
            hyperparameters=hp,
            allow_new_entries=False,
        )
        tuner.search(
            x=TRAIN_INPUTS,
            y=TRAIN_TARGETS,
            epochs=2,
            validation_data=(VAL_INPUTS, VAL_TARGETS),
        )


def test_restricted_space_using_defaults(tmp_path):
    hp = keras_tuner.HyperParameters()
    hp.Int("num_layers", 1, 5, 1, default=2)
    hp.Choice("learning_rate", [0.01, 0.001, 0.0001])

    tuner = keras_tuner.tuners.RandomSearch(
        build_model,
        objective="val_accuracy",
        max_trials=4,
        directory=tmp_path,
        hyperparameters=hp,
        allow_new_entries=True,
        tune_new_entries=False,
    )

    assert len(tuner.oracle.hyperparameters.space) == 2
    new_lr = [
        p
        for p in tuner.oracle.hyperparameters.space
        if p.name == "learning_rate"
    ][0]
    assert new_lr.values == [0.01, 0.001, 0.0001]

    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=1,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    assert len(tuner.oracle.trials) == 4
    assert len(tuner.oracle.hyperparameters.space) == 2  # Nothing added
    for trial in tuner.oracle.trials.values():
        # Trials get default values but don't pass these on to the oracle.
        assert len(trial.hyperparameters.space) >= 2


def test_restricted_space_with_custom_defaults(tmp_path):
    hp = keras_tuner.HyperParameters()
    hp.Int("num_layers", 1, 3, 1, default=2)
    hp.Choice("learning_rate", [0.01, 0.001, 0.0001])
    hp.Fixed("units_0", 4)
    hp.Fixed("units_1", 3)

    tuner = keras_tuner.tuners.RandomSearch(
        build_model,
        objective="val_accuracy",
        max_trials=4,
        directory=tmp_path,
        hyperparameters=hp,
        allow_new_entries=True,
        tune_new_entries=False,
    )

    assert len(tuner.oracle.hyperparameters.space) == 4
    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=1,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    assert len(tuner.oracle.trials) == 4


def test_reparameterized_space(tmp_path):
    hp = keras_tuner.HyperParameters()
    hp.Int("num_layers", 1, 3, 1, default=3)
    hp.Choice("learning_rate", [0.01, 0.001, 0.0001])

    tuner = keras_tuner.tuners.RandomSearch(
        build_model,
        seed=1337,
        objective="val_accuracy",
        max_trials=4,
        directory=tmp_path,
        hyperparameters=hp,
        allow_new_entries=True,
        tune_new_entries=True,
    )

    # Initial build model adds to the space.
    assert len(tuner.oracle.hyperparameters.space) == 5
    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=1,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    assert len(tuner.oracle.trials) == 4
    assert len(tuner.oracle.hyperparameters.space) == 5


def test_get_best_models(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        build_model, objective="val_accuracy", max_trials=4, directory=tmp_path
    )

    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=2,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    models = tuner.get_best_models(2)
    assert len(models) == 2
    assert isinstance(models[0], keras.Model)
    assert isinstance(models[1], keras.Model)


def test_saving_and_reloading(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        build_model,
        objective="val_accuracy",
        max_trials=4,
        executions_per_trial=2,
        directory=tmp_path,
    )

    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=2,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    new_tuner = keras_tuner.tuners.RandomSearch(
        build_model,
        objective="val_accuracy",
        max_trials=4,
        executions_per_trial=2,
        directory=tmp_path,
    )
    new_tuner.reload()

    assert len(new_tuner.oracle.trials) == 4

    new_tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=2,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )


def test_subclass_model(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        build_subclass_model,
        objective="val_accuracy",
        max_trials=2,
        directory=tmp_path,
    )

    tuner.search_space_summary()

    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=2,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    tuner.results_summary()
    assert len(tuner.oracle.trials) == 2


def test_subclass_model_loading(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        build_subclass_model,
        objective="val_accuracy",
        max_trials=2,
        directory=tmp_path,
    )

    tuner.search_space_summary()

    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=2,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    best_trial_score = tuner.oracle.get_best_trials()[0].score

    best_model = tuner.get_best_models()[0]
    best_model_score = best_model.evaluate(VAL_INPUTS, VAL_TARGETS)[1]

    assert best_model_score == best_trial_score


def test_update_trial(tmp_path):
    # Test stop the oracle in update_trial.
    class MyOracle(keras_tuner.Oracle):
        def populate_space(self, _):
            values = {
                p.name: p.random_sample() for p in self.hyperparameters.space
            }
            return {"values": values, "status": "RUNNING"}

        def update_trial(self, trial_id, metrics, step=0):
            super().update_trial(trial_id, metrics, step)
            trial = self.trials[trial_id]
            trial.status = "STOPPED"
            return trial.status

    my_oracle = MyOracle(objective="val_accuracy", max_trials=2)
    tuner = keras_tuner.Tuner(
        oracle=my_oracle, hypermodel=build_model, directory=tmp_path
    )
    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=5,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    assert len(my_oracle.trials) == 2

    for trial in my_oracle.trials.values():
        # Test that early stopping worked.
        assert len(trial.metrics.get_history("val_accuracy")) == 1


def test_tunable_false_hypermodel(tmp_path):
    def build_model(hp):
        input_shape = (256, 256, 3)
        inputs = tf.keras.Input(shape=input_shape)

        with hp.name_scope("xception"):
            # Tune the pooling of Xception by supplying the search space
            # beforehand.
            hp.Choice("pooling", ["avg", "max"])
            xception = keras_tuner.applications.HyperXception(
                include_top=False, input_shape=input_shape, tunable=False
            ).build(hp)
        x = xception(inputs)

        x = tf.keras.layers.Dense(
            hp.Int("hidden_units", 50, 100, step=10), activation="relu"
        )(x)
        outputs = tf.keras.layers.Dense(NUM_CLASSES, activation="softmax")(x)

        model = tf.keras.Model(inputs, outputs)

        optimizer = tf.keras.optimizers.get(
            hp.Choice("optimizer", ["adam", "sgd"])
        )
        optimizer.learning_rate = hp.Float(
            "learning_rate", 1e-4, 1e-2, sampling="log"
        )

        model.compile(optimizer, loss="sparse_categorical_crossentropy")
        return model

    tuner = keras_tuner.RandomSearch(
        objective="val_loss",
        hypermodel=build_model,
        max_trials=4,
        directory=tmp_path,
    )

    x = np.random.random(size=(2, 256, 256, 3))
    y = np.random.randint(0, NUM_CLASSES, size=(2,))

    tuner.search(x, y, validation_data=(x, y), batch_size=2)

    hps = tuner.oracle.get_space()
    assert "xception/pooling" in hps
    assert "hidden_units" in hps
    assert "optimizer" in hps
    assert "learning_rate" in hps

    # Make sure no HPs from building xception were added.
    assert len(hps.space) == 4


def test_get_best_hyperparameters(tmp_path):
    hp1 = keras_tuner.HyperParameters()
    hp1.Fixed("a", 1)
    trial1 = keras_tuner.engine.trial.Trial(hyperparameters=hp1)
    trial1.status = "COMPLETED"
    trial1.score = 10

    hp2 = keras_tuner.HyperParameters()
    hp2.Fixed("a", 2)
    trial2 = keras_tuner.engine.trial.Trial(hyperparameters=hp2)
    trial2.status = "COMPLETED"
    trial2.score = 9

    tuner = keras_tuner.RandomSearch(
        objective="val_accuracy",
        hypermodel=build_model,
        max_trials=2,
        directory=tmp_path,
    )

    tuner.oracle.trials = {trial1.trial_id: trial1, trial2.trial_id: trial2}

    hps = tuner.get_best_hyperparameters()[0]
    assert hps["a"] == 1


def test_reloading_error_message(tmp_path):
    shared_dir = tmp_path

    tuner = keras_tuner.tuners.RandomSearch(
        build_model,
        objective="val_accuracy",
        max_trials=2,
        executions_per_trial=3,
        directory=shared_dir,
    )

    tuner.search(
        x=TRAIN_INPUTS,
        y=TRAIN_TARGETS,
        epochs=2,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    with pytest.raises(RuntimeError, match="pass `overwrite=True`"):
        keras_tuner.tuners.BayesianOptimization(
            build_model,
            objective="val_accuracy",
            max_trials=2,
            executions_per_trial=3,
            directory=shared_dir,
        )


def test_search_logging_verbosity(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        build_model,
        objective="val_accuracy",
        max_trials=2,
        executions_per_trial=3,
        directory=tmp_path,
    )

    with patch("sys.stdout", new=StringIO()) as output:
        tuner.search(
            x=TRAIN_INPUTS,
            y=TRAIN_TARGETS,
            epochs=2,
            validation_data=(VAL_INPUTS, VAL_TARGETS),
            verbose=0,
        )
        assert output.getvalue().strip() == ""


def test_convert_hyperparams_to_hparams():
    def _check_hparams_equal(hp1, hp2):
        assert (
            hparams_api.hparams_pb(hp1, start_time_secs=0).SerializeToString()
            == hparams_api.hparams_pb(
                hp2, start_time_secs=0
            ).SerializeToString()
        )

    hps = keras_tuner.engine.hyperparameters.HyperParameters()
    hps.Choice("learning_rate", [1e-4, 1e-3, 1e-2])
    hparams = keras_tuner.engine.tuner_utils.convert_hyperparams_to_hparams(hps)
    _check_hparams_equal(
        hparams,
        {
            hparams_api.HParam(
                "learning_rate", hparams_api.Discrete([1e-4, 1e-3, 1e-2])
            ): 1e-4
        },
    )

    hps = keras_tuner.engine.hyperparameters.HyperParameters()
    hps.Int("units", min_value=2, max_value=16)
    hparams = keras_tuner.engine.tuner_utils.convert_hyperparams_to_hparams(hps)
    _check_hparams_equal(
        hparams,
        {hparams_api.HParam("units", hparams_api.IntInterval(2, 16)): 2},
    )

    hps = keras_tuner.engine.hyperparameters.HyperParameters()
    hps.Int("units", min_value=32, max_value=128, step=32)
    hparams = keras_tuner.engine.tuner_utils.convert_hyperparams_to_hparams(hps)
    _check_hparams_equal(
        hparams,
        {
            hparams_api.HParam(
                "units", hparams_api.Discrete([32, 64, 96, 128])
            ): 32
        },
    )

    hps = keras_tuner.engine.hyperparameters.HyperParameters()
    hps.Float("learning_rate", min_value=0.5, max_value=1.25, step=0.25)
    hparams = keras_tuner.engine.tuner_utils.convert_hyperparams_to_hparams(hps)
    _check_hparams_equal(
        hparams,
        {
            hparams_api.HParam(
                "learning_rate", hparams_api.Discrete([0.5, 0.75, 1.0, 1.25])
            ): 0.5
        },
    )

    hps = keras_tuner.engine.hyperparameters.HyperParameters()
    hps.Float("learning_rate", min_value=1e-4, max_value=1e-1)
    hparams = keras_tuner.engine.tuner_utils.convert_hyperparams_to_hparams(hps)
    _check_hparams_equal(
        hparams,
        {
            hparams_api.HParam(
                "learning_rate", hparams_api.RealInterval(1e-4, 1e-1)
            ): 1e-4
        },
    )

    hps = keras_tuner.engine.hyperparameters.HyperParameters()
    hps.Float("theta", min_value=0.0, max_value=1.57)
    hps.Float("r", min_value=0.0, max_value=1.0)
    hparams = keras_tuner.engine.tuner_utils.convert_hyperparams_to_hparams(hps)
    expected_hparams = {
        hparams_api.HParam("theta", hparams_api.RealInterval(0.0, 1.57)): 0.0,
        hparams_api.HParam("r", hparams_api.RealInterval(0.0, 1.0)): 0.0,
    }
    hparams_repr_list = [repr(hparams[x]) for x in hparams.keys()]
    expected_hparams_repr_list = [
        repr(expected_hparams[x]) for x in expected_hparams
    ]

    assert sorted(hparams_repr_list) == sorted(expected_hparams_repr_list)

    hps = keras_tuner.engine.hyperparameters.HyperParameters()
    hps.Boolean("has_beta")
    hparams = keras_tuner.engine.tuner_utils.convert_hyperparams_to_hparams(hps)
    _check_hparams_equal(
        hparams,
        {
            hparams_api.HParam(
                "has_beta", hparams_api.Discrete([True, False])
            ): False
        },
    )

    hps = keras_tuner.engine.hyperparameters.HyperParameters()
    hps.Fixed("beta", 0.1)
    hparams = keras_tuner.engine.tuner_utils.convert_hyperparams_to_hparams(hps)
    _check_hparams_equal(
        hparams, {hparams_api.HParam("beta", hparams_api.Discrete([0.1])): 0.1}
    )

    hps = keras_tuner.engine.hyperparameters.HyperParameters()
    hps.Fixed("type", "WIDE_AND_DEEP")
    hparams = keras_tuner.engine.tuner_utils.convert_hyperparams_to_hparams(hps)
    _check_hparams_equal(
        hparams,
        {
            hparams_api.HParam(
                "type", hparams_api.Discrete(["WIDE_AND_DEEP"])
            ): "WIDE_AND_DEEP"
        },
    )

    hps = keras_tuner.engine.hyperparameters.HyperParameters()
    hps.Fixed("condition", True)
    hparams = keras_tuner.engine.tuner_utils.convert_hyperparams_to_hparams(hps)
    _check_hparams_equal(
        hparams,
        {hparams_api.HParam("condition", hparams_api.Discrete([True])): True},
    )

    hps = keras_tuner.engine.hyperparameters.HyperParameters()
    hps.Fixed("num_layers", 2)
    hparams = keras_tuner.engine.tuner_utils.convert_hyperparams_to_hparams(hps)
    _check_hparams_equal(
        hparams,
        {hparams_api.HParam("num_layers", hparams_api.Discrete([2])): 2},
    )


def test_tuning_correctness(tmp_path):
    tuner = keras_tuner.Tuner(
        oracle=keras_tuner.tuners.randomsearch.RandomSearchOracle(
            objective="loss", max_trials=2, seed=1337
        ),
        hypermodel=MockHyperModel(),
        directory=tmp_path,
    )
    tuner.search()
    assert len(tuner.oracle.trials) == 2

    m0_epochs = [float(np.average(x)) for x in MockHyperModel.mode_0]
    m1_epochs = [float(np.average(x)) for x in MockHyperModel.mode_1]

    # Score tracking correctness
    first_trial, second_trial = sorted(
        tuner.oracle.trials.values(), key=lambda t: t.score
    )
    assert first_trial.score == min(m0_epochs)
    assert second_trial.score == min(m1_epochs)
    assert tuner.oracle.get_best_trials(1)[0].trial_id == first_trial.trial_id


def assert_found_best_score(
    tmp_path, hypermodel, tuner_class=keras_tuner.Tuner
):
    tuner = tuner_class(
        oracle=keras_tuner.tuners.randomsearch.RandomSearchOracle(
            objective="loss", max_trials=2, seed=1337
        ),
        hypermodel=hypermodel,
        directory=tmp_path,
    )
    tuner.search(callbacks=[])
    assert tuner.oracle.get_best_trials(1)[0].score == 3.0


def test_hypermodel_fit_return_a_dict(tmp_path):
    class MyHyperModel(MockHyperModel):
        def fit(self, hp, model, *args, **kwargs):
            history = super().fit(hp, model, *args, **kwargs)
            return {
                "loss": min(history.history["loss"]),
                "other_metric": np.random.rand(),
            }

    assert_found_best_score(tmp_path, MyHyperModel())


def test_hypermodel_fit_return_a_float(tmp_path):
    class MyHyperModel(MockHyperModel):
        def fit(self, hp, model, *args, **kwargs):
            history = super().fit(hp, model, *args, **kwargs)
            return min(history.history["loss"])

    assert_found_best_score(tmp_path, MyHyperModel())


def test_hypermodel_fit_return_an_int(tmp_path):
    class MyHyperModel(MockHyperModel):
        def fit(self, hp, model, *args, **kwargs):
            history = super().fit(hp, model, *args, **kwargs)
            return int(min(history.history["loss"]))

    assert_found_best_score(tmp_path, MyHyperModel())


def test_run_trial_return_none_without_update_trial(tmp_path):
    class MyTuner(keras_tuner.Tuner):
        def run_trial(self, trial, *fit_args, **fit_kwargs):
            self.hypermodel.build(trial.hyperparameters).fit(
                *fit_args, **fit_kwargs
            )

    with pytest.raises(
        errors.FatalTypeError,
        match="Did you forget",
    ):
        assert_found_best_score(tmp_path, MockHyperModel(), MyTuner)


def test_run_trial_return_none_with_update_trial(tmp_path):
    class MyTuner(keras_tuner.Tuner):
        def run_trial(self, trial, *fit_args, **fit_kwargs):
            history = self.hypermodel.build(trial.hyperparameters).fit(
                *fit_args, **fit_kwargs
            )
            self.oracle.update_trial(
                trial.trial_id, {"loss": min(history.history["loss"])}
            )

    with pytest.deprecated_call(match="Please remove the call"):
        assert_found_best_score(tmp_path, MockHyperModel(), MyTuner)


def test_run_trial_return_history(tmp_path):
    class MyTuner(keras_tuner.Tuner):
        def run_trial(self, trial, *fit_args, **fit_kwargs):
            return self.hypermodel.build(trial.hyperparameters).fit(
                *fit_args, **fit_kwargs
            )

    assert_found_best_score(tmp_path, MockHyperModel(), MyTuner)


def test_run_trial_return_a_dict(tmp_path):
    class MyTuner(keras_tuner.Tuner):
        def run_trial(self, trial, *fit_args, **fit_kwargs):
            history = self.hypermodel.build(trial.hyperparameters).fit(
                *fit_args, **fit_kwargs
            )
            return {"loss": min(history.history["loss"])}

    assert_found_best_score(tmp_path, MockHyperModel(), MyTuner)


def test_run_trial_return_a_float(tmp_path):
    class MyTuner(keras_tuner.Tuner):
        def run_trial(self, trial, *fit_args, **fit_kwargs):
            history = self.hypermodel.build(trial.hyperparameters).fit(
                *fit_args, **fit_kwargs
            )
            return min(history.history["loss"])

    assert_found_best_score(tmp_path, MockHyperModel(), MyTuner)


def test_run_trial_return_float_list(tmp_path):
    class MyTuner(keras_tuner.Tuner):
        def run_trial(self, trial, *fit_args, **fit_kwargs):
            ret = []
            for _ in range(3):
                history = self.hypermodel.build(trial.hyperparameters).fit(
                    *fit_args, **fit_kwargs
                )
                ret.append(min(history.history["loss"]))
            return ret

    assert_found_best_score(tmp_path, MockHyperModel(), MyTuner)


def test_tuner_errors(tmp_path):
    # invalid oracle
    with pytest.raises(
        ValueError,
        match="Expected `oracle` argument to be an instance of `Oracle`",
    ):
        tuner_module.Tuner(
            oracle="invalid", hypermodel=build_model, directory=tmp_path
        )
    # invalid hypermodel
    with pytest.raises(
        ValueError, match="`hypermodel` argument should be either"
    ):
        tuner_module.Tuner(
            oracle=keras_tuner.tuners.randomsearch.RandomSearchOracle(
                objective="val_accuracy", max_trials=3
            ),
            hypermodel="build_model",
            directory=tmp_path,
        )
    # oversize model
    with pytest.raises(RuntimeError, match="Oversized model"):
        tuner = tuner_module.Tuner(
            oracle=keras_tuner.tuners.randomsearch.RandomSearchOracle(
                objective="val_accuracy", max_trials=3
            ),
            hypermodel=build_model,
            max_model_size=4,
            directory=tmp_path,
        )
        tuner.search(
            TRAIN_INPUTS,
            TRAIN_TARGETS,
            validation_data=(VAL_INPUTS, VAL_TARGETS),
        )
    # TODO: test no optimizer


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="TPUStrategy only exists in TF2.3+.",
)
def test_metric_direction_inferred_from_objective(tmp_path):
    oracle = keras_tuner.tuners.randomsearch.RandomSearchOracle(
        objective=keras_tuner.Objective("a", "max"), max_trials=1
    )
    oracle._set_project_dir(tmp_path, "untitled_project")
    trial = oracle.create_trial("tuner0")
    oracle.update_trial(trial.trial_id, {"a": 1})
    trial = oracle.get_trial(trial.trial_id)
    assert trial.metrics.get_direction("a") == "max"

    oracle = keras_tuner.tuners.randomsearch.RandomSearchOracle(
        objective=keras_tuner.Objective("a", "min"), max_trials=1
    )
    oracle._set_project_dir(tmp_path, "untitled_project2")
    trial = oracle.create_trial("tuner0")
    oracle.update_trial(trial.trial_id, {"a": 1})
    trial = oracle.get_trial(trial.trial_id)
    assert trial.metrics.get_direction("a") == "min"


def test_overwrite_true(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        hypermodel=build_model,
        objective="val_accuracy",
        max_trials=2,
        directory=tmp_path,
    )
    tuner.search(
        TRAIN_INPUTS, TRAIN_TARGETS, validation_data=(VAL_INPUTS, VAL_TARGETS)
    )
    assert len(tuner.oracle.trials) == 2

    new_tuner = keras_tuner.tuners.RandomSearch(
        hypermodel=build_model,
        objective="val_accuracy",
        max_trials=2,
        directory=tmp_path,
        overwrite=True,
    )
    assert len(new_tuner.oracle.trials) == 0


def test_correct_display_trial_number(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        hypermodel=build_model,
        objective="val_accuracy",
        max_trials=2,
        directory=tmp_path,
    )
    tuner.search(
        TRAIN_INPUTS, TRAIN_TARGETS, validation_data=(VAL_INPUTS, VAL_TARGETS)
    )
    new_tuner = keras_tuner.tuners.RandomSearch(
        hypermodel=build_model,
        objective="val_accuracy",
        max_trials=6,
        directory=tmp_path,
        overwrite=False,
    )
    new_tuner.search(
        TRAIN_INPUTS, TRAIN_TARGETS, validation_data=(VAL_INPUTS, VAL_TARGETS)
    )
    assert len(new_tuner.oracle.trials) == new_tuner._display.trial_number


def test_error_on_unknown_objective_direction(tmp_path):
    with pytest.raises(
        ValueError, match="Could not infer optimization direction"
    ):
        keras_tuner.tuners.RandomSearch(
            hypermodel=build_model,
            objective="custom_metric",
            max_trials=2,
            directory=tmp_path,
        )


def test_callbacks_run_each_execution(tmp_path):
    callback_instances = set()

    class LoggingCallback(keras.callbacks.Callback):
        def on_train_begin(self, logs):
            callback_instances.add(id(self))

    logging_callback = LoggingCallback()
    tuner = keras_tuner.tuners.RandomSearch(
        hypermodel=build_model,
        objective="val_accuracy",
        max_trials=2,
        executions_per_trial=3,
        directory=tmp_path,
    )
    tuner.search(
        TRAIN_INPUTS,
        TRAIN_TARGETS,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
        callbacks=[logging_callback],
    )

    # Unknown reason cause the callback to run 5 times sometime.
    # Make 5 & 6 both pass the test before found the reason.
    assert len(callback_instances) in {5, 6}


def test_build_and_fit_model(tmp_path):
    class MyTuner(keras_tuner.tuners.RandomSearch):
        def _build_and_fit_model(self, trial, *args, **kwargs):
            self.was_called = True
            return super()._build_and_fit_model(trial, *args, **kwargs)

    tuner = MyTuner(
        hypermodel=build_model,
        objective="val_accuracy",
        max_trials=2,
        executions_per_trial=3,
        directory=tmp_path,
    )

    tuner.run_trial(
        tuner.oracle.create_trial("tuner0"),
        TRAIN_INPUTS,
        TRAIN_TARGETS,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    assert tuner.was_called


def test_build_and_fit_model_in_tuner(tmp_path):
    class MyTuner(tuner_module.Tuner):
        def _build_and_fit_model(self, trial, *args, **kwargs):
            self.was_called = True
            return super()._build_and_fit_model(trial, *args, **kwargs)

    tuner = MyTuner(
        oracle=keras_tuner.tuners.randomsearch.RandomSearchOracle(
            objective="val_loss",
            max_trials=2,
        ),
        hypermodel=build_model,
        directory=tmp_path,
    )

    tuner.run_trial(
        tuner.oracle.create_trial("tuner0"),
        TRAIN_INPUTS,
        TRAIN_TARGETS,
        validation_data=(VAL_INPUTS, VAL_TARGETS),
    )

    assert tuner.was_called


def test_init_build_all_hps_in_all_conditions(tmp_path):
    class ConditionalHyperModel(MockHyperModel):
        def build(self, hp):
            model_type = hp.Choice("model_type", ["cnn", "mlp"])
            with hp.conditional_scope("model_type", ["cnn"]):
                if model_type == "cnn":
                    sub_cnn = hp.Choice("sub_cnn", ["a", "b"])
                    with hp.conditional_scope("sub_cnn", ["a"]):
                        if sub_cnn == "a":
                            hp.Int("n_filters_a", 2, 4)
                    with hp.conditional_scope("sub_cnn", ["b"]):
                        if sub_cnn == "b":
                            hp.Int("n_filters_b", 6, 8)
            with hp.conditional_scope("model_type", ["mlp"]):
                if model_type == "mlp":
                    sub_mlp = hp.Choice("sub_mlp", ["a", "b"])
                    with hp.conditional_scope("sub_mlp", ["a"]):
                        if sub_mlp == "a":
                            hp.Int("n_units_a", 2, 4)
                    with hp.conditional_scope("sub_mlp", ["b"]):
                        if sub_mlp == "b":
                            hp.Int("n_units_b", 6, 8)
            more_block = hp.Boolean("more_block", default=False)
            with hp.conditional_scope("more_block", [True]):
                if more_block:
                    hp.Int("new_block_hp", 1, 3)
            return super().build(hp)

    def name_in_hp(name, hp):
        return any(name == single_hp.name for single_hp in hp.space)

    class MyTuner(tuner_module.Tuner):
        def _populate_initial_space(self):
            super()._populate_initial_space()
            hp = self.oracle.hyperparameters
            assert name_in_hp("model_type", hp)
            assert name_in_hp("sub_cnn", hp)
            assert name_in_hp("n_filters_a", hp)
            assert name_in_hp("n_filters_b", hp)
            assert name_in_hp("sub_mlp", hp)
            assert name_in_hp("n_units_a", hp)
            assert name_in_hp("n_units_b", hp)
            assert name_in_hp("more_block", hp)
            assert name_in_hp("new_block_hp", hp)

    MyTuner(
        oracle=keras_tuner.tuners.randomsearch.RandomSearchOracle(
            objective="loss", max_trials=2, seed=1337
        ),
        hypermodel=ConditionalHyperModel(),
        directory=tmp_path,
    )


def test_populate_initial_space_with_hp_parent_arg(tmp_path):
    def build_model(hp):
        hp.Boolean("parent", default=True)
        hp.Boolean(
            "child",
            parent_name="parent",
            parent_values=[False],
        )
        return keras.Sequential()

    keras_tuner.RandomSearch(
        build_model,
        objective="val_accuracy",
        directory=tmp_path,
        max_trials=1,
    )


def test_populate_initial_space_with_declare_hp(tmp_path):
    class MyHyperModel(keras_tuner.HyperModel):
        def declare_hyperparameters(self, hp):
            hp.Boolean("bool")

        def build(self, hp):
            return keras.Sequential()

    keras_tuner.RandomSearch(
        MyHyperModel(),
        objective="val_accuracy",
        directory=tmp_path,
        max_trials=1,
    )


def test_build_did_not_return_keras_model(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        hypermodel=lambda hp: None,
        objective="val_accuracy",
        directory=tmp_path,
    )
    with pytest.raises(
        errors.FatalTypeError,
        match="Expected the model-building function",
    ):
        tuner.search()


def test_callback_cannot_be_deep_copied(tmp_path):
    tuner = keras_tuner.tuners.RandomSearch(
        hypermodel=lambda hp: keras.Sequential(),
        objective="val_accuracy",
        directory=tmp_path,
    )
    with pytest.raises(
        errors.FatalValueError,
        match="All callbacks used during a search should be deep-copyable",
    ):
        tuner.search(callbacks=[keras_tuner])



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OracleClient class."""

import os

import grpc

from keras_tuner import protos
from keras_tuner.engine import hyperparameters as hp_module
from keras_tuner.engine import trial as trial_module


class OracleClient:
    """Wraps an `Oracle` on a worker to send requests to the chief."""

    def __init__(self, oracle):
        self._oracle = oracle

        ip_addr = os.environ["KERASTUNER_ORACLE_IP"]
        port = os.environ["KERASTUNER_ORACLE_PORT"]
        channel = grpc.insecure_channel(f"{ip_addr}:{port}")
        self.stub = protos.get_service_grpc().OracleStub(channel)
        self.tuner_id = os.environ["KERASTUNER_TUNER_ID"]

        # In multi-worker mode, only the chief of each cluster should report
        # results to the chief Oracle.
        self.multi_worker = False
        self.should_report = True

    def __getattr__(self, name):
        whitelisted_attrs = {
            "objective",
            "max_trials",
            "allow_new_entries",
            "tune_new_entries",
        }
        if name in whitelisted_attrs:
            return getattr(self._oracle, name)
        raise AttributeError(f'`OracleClient` object has no attribute "{name}"')

    def get_space(self):
        response = self.stub.GetSpace(
            protos.get_service().GetSpaceRequest(), wait_for_ready=True
        )
        return hp_module.HyperParameters.from_proto(response.hyperparameters)

    def update_space(self, hyperparameters):
        if self.should_report:
            self.stub.UpdateSpace(
                protos.get_service().UpdateSpaceRequest(
                    hyperparameters=hyperparameters.to_proto()
                ),
                wait_for_ready=True,
            )

    def create_trial(self, tuner_id):
        response = self.stub.CreateTrial(
            protos.get_service().CreateTrialRequest(tuner_id=tuner_id),
            wait_for_ready=True,
        )
        return trial_module.Trial.from_proto(response.trial)

    def update_trial(self, trial_id, metrics, step=0):
        # TODO: support early stopping in multi-worker.
        if self.should_report:
            response = self.stub.UpdateTrial(
                protos.get_service().UpdateTrialRequest(
                    trial_id=trial_id, metrics=metrics, step=step
                ),
                wait_for_ready=True,
            )
            if not self.multi_worker:
                return trial_module.Trial.from_proto(response.trial)
        return trial_module.Trial(self.get_space(), status="RUNNING")

    def end_trial(self, trial):
        if self.should_report:
            self.stub.EndTrial(
                protos.get_service().EndTrialRequest(trial=trial.to_proto()),
                wait_for_ready=True,
            )

    def get_trial(self, trial_id):
        response = self.stub.GetTrial(
            protos.get_service().GetTrialRequest(trial_id=trial_id),
            wait_for_ready=True,
        )
        return trial_module.Trial.from_proto(response.trial)

    def get_best_trials(self, num_trials=1):
        response = self.stub.GetBestTrials(
            protos.get_service().GetBestTrialsRequest(num_trials=num_trials),
            wait_for_ready=True,
        )
        return [
            trial_module.Trial.from_proto(trial) for trial in response.trials
        ]



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_tuner import protos
from keras_tuner.api_export import keras_tuner_export
from keras_tuner.engine import conditions as conditions_mod
from keras_tuner.engine.hyperparameters import hyperparameter


@keras_tuner_export("keras_tuner.engine.hyperparameters.Boolean")
class Boolean(hyperparameter.HyperParameter):
    """Choice between True and False.

    Args:
        name: A string. the name of parameter. Must be unique for each
            `HyperParameter` instance in the search space.
        default: Boolean, the default value to return for the parameter.
            If unspecified, the default value will be False.
    """

    def __init__(self, name, default=False, **kwargs):
        super().__init__(name=name, default=default, **kwargs)
        if default not in {True, False}:
            raise ValueError(
                "`default` must be a Python boolean. "
                f"You passed: default={default}"
            )

    def __repr__(self):
        return f'Boolean(name: "{self.name}", default: {self.default})'

    @property
    def values(self):
        return (True, False)

    def prob_to_value(self, prob):
        return bool(prob >= 0.5)

    def value_to_prob(self, value):
        # Center the value in its probability bucket.
        return 0.75 if value else 0.25

    @classmethod
    def from_proto(cls, proto):
        conditions = [
            conditions_mod.Condition.from_proto(c) for c in proto.conditions
        ]
        return cls(
            name=proto.name, default=proto.default, conditions=conditions
        )

    def to_proto(self):
        return protos.get_proto().Boolean(
            name=self.name,
            default=self.default,
            conditions=[c.to_proto() for c in self.conditions],
        )



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
from concurrent import futures

import grpc

from keras_tuner import protos
from keras_tuner.engine import hyperparameters as hp_module
from keras_tuner.engine import trial as trial_module


class OracleServicer(protos.get_service_grpc().OracleServicer):
    def __init__(self, oracle):
        self.oracle = oracle
        self.stop_triggered = False

    def GetSpace(self, request, context):
        hps = self.oracle.get_space()
        return protos.get_service().GetSpaceResponse(
            hyperparameters=hps.to_proto()
        )

    def UpdateSpace(self, request, context):
        hps = hp_module.HyperParameters.from_proto(request.hyperparameters)
        self.oracle.update_space(hps)
        return protos.get_service().UpdateSpaceResponse()

    def CreateTrial(self, request, context):
        trial = self.oracle.create_trial(request.tuner_id)
        if trial.status == trial_module.TrialStatus.STOPPED:
            self.stop_triggered = True
        return protos.get_service().CreateTrialResponse(trial=trial.to_proto())

    def UpdateTrial(self, request, context):
        trial = self.oracle.update_trial(
            request.trial_id, request.metrics, step=request.step
        )
        return protos.get_service().UpdateTrialResponse(trial=trial.to_proto())

    def EndTrial(self, request, context):
        trial = trial_module.Trial.from_proto(request.trial)
        self.oracle.end_trial(trial)
        return protos.get_service().EndTrialResponse()

    def GetTrial(self, request, context):
        trial = self.oracle.get_trial(request.trial_id)
        return protos.get_service().GetTrialResponse(trial=trial.to_proto())

    def GetBestTrials(self, request, context):
        trials = self.oracle.get_best_trials(request.num_trials)
        return protos.get_service().GetBestTrialsResponse(
            trials=[trial.to_proto() for trial in trials]
        )


def start_server(oracle):
    """Starts the `OracleServicer` used to manage distributed requests."""
    ip_addr = os.environ["KERASTUNER_ORACLE_IP"]
    port = os.environ["KERASTUNER_ORACLE_PORT"]
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    oracle_servicer = OracleServicer(oracle)
    protos.get_service_grpc().add_OracleServicer_to_server(
        oracle_servicer, server
    )
    server.add_insecure_port(f"{ip_addr}:{port}")
    server.start()
    while True:
        # The server does not block otherwise.
        time.sleep(30)

        if oracle_servicer.stop_triggered:
            while oracle.ongoing_trials:
                time.sleep(10)  # pragma: no cover

            print(
                "Oracle server on chief is exiting in 40s."
                "The chief will go on with post-search code."
            )
            # Wait for another 30s for all the subsequent calls from workers to
            # come
            time.sleep(30)
            server.stop(10)
            break



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from keras_tuner import protos
from keras_tuner.engine import hyperparameters as hp_module
from keras_tuner.engine.hyperparameters import hp_types


def test_int_sampling_arg():
    i = hp_module.Int("i", 0, 10, sampling="linear")
    i = hp_module.Int.from_config(i.get_config())
    assert i.sampling == "linear"

    with pytest.raises(ValueError, match="sampling must be one of"):
        hp_module.Int("j", 0, 10, sampling="invalid")

    with pytest.raises(
        ValueError,
        match="min_value 1 is greater than the max_value 0",
    ):
        hp_module.Int("k", 1, 0, sampling="linear")

    with pytest.raises(
        ValueError,
        match="min_value 1 is greater than the max_value 0",
    ):
        hp_module.Int("k", 1, 0, sampling="linear")

    with pytest.raises(
        ValueError,
        match="does not support negative values",
    ):
        hp_module.Int("k", -10, -1, sampling="log")

    with pytest.raises(
        ValueError,
        match="For HyperParameters.Int\(name='k'\), expected step > 1",
    ):
        hp_module.Int("k", 1, 10, step=1, sampling="log")


def test_int():
    rg = hp_module.Int("rg", min_value=5, max_value=9, step=1, default=6)
    rg = hp_module.Int.from_config(rg.get_config())
    assert rg.default == 6
    assert 5 <= rg.random_sample() <= 9
    assert isinstance(rg.random_sample(), int)
    assert rg.random_sample(123) == rg.random_sample(123)
    assert abs(rg.value_to_prob(6) - 0.3) < 1e-4
    # No default
    rg = hp_module.Int("rg", min_value=5, max_value=9, step=1)
    assert rg.default == 5


def test_int_log_with_step():
    rg = hp_module.Int("rg", min_value=2, max_value=32, step=2, sampling="log")
    for _ in range(10):
        assert rg.random_sample() in [2, 4, 8, 16, 32]
    assert abs(rg.value_to_prob(4) - 0.3) < 1e-4
    assert rg.prob_to_value(0.3) == 4


def test_int_log_without_step():
    rg = hp_module.Int("rg", min_value=2, max_value=32, sampling="log")
    assert rg.prob_to_value(rg.value_to_prob(4)) == 4


def test_int_proto():
    hp = hp_module.Int("a", 1, 100, sampling="log")
    proto = hp.to_proto()
    assert proto.name == "a"
    assert proto.min_value == 1
    assert proto.max_value == 100
    assert proto.sampling == protos.get_proto().Sampling.LOG
    # Proto stores the implicit default.
    assert proto.default == 1
    assert proto.step == 0

    new_hp = hp_module.Int.from_proto(proto)
    assert new_hp._default == 1
    # Pop the implicit default for comparison purposes.
    new_hp._default = None
    assert new_hp.get_config() == hp.get_config()


def test_int_raise_error_with_float_min_value():
    with pytest.raises(ValueError, match="must be an int"):
        hp_module.Int("j", 0.5, 10)


def test_repr_int_is_str():
    assert "name: 'j'" in repr(hp_module.Int("j", 1, 10))


def test_serialize_deserialize_int():
    hp = hp_module.Int("j", 1, 10)
    new_hp = hp_module.deserialize(hp_module.serialize(hp))
    assert repr(hp) == repr(new_hp)

    hp = hp_module.Int("j", 1, 10)
    new_hp = hp_types.deserialize(hp_types.serialize(hp))
    assert repr(hp) == repr(new_hp)


def test_int_values_property_with_step():
    assert list(hp_module.Int("int", 2, 8, 2).values) == [2, 4, 6, 8]
    assert isinstance(list(hp_module.Int("int", 2, 8, 2).values)[0], int)
    assert list(hp_module.Int("int", 2, 8, 2, sampling="log").values) == [
        2,
        4,
        8,
    ]


def test_int_values_property_without_step():
    assert list(hp_module.Int("int", 2, 4).values) == [2, 3, 4]
    assert list(hp_module.Int("int", 2, 20).values) == list(range(2, 21))
    assert len(list(hp_module.Int("int", 2, 1024, sampling="log").values)) == 10


def test_sampling_none_is_linear():
    # This is for backward compatibility
    assert hp_module.Int("int", 2, 4, sampling=None).sampling == "linear"



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""KerasTuner protos."""

from contextlib import contextmanager

# v3 is generated with protobuf==3.20.3 and grpcio-tools==1.48.0
# v4 is generated with protobuf==4.22.1 and grpcio-tools==1.53.0
try:
    from keras_tuner.protos import v4 as protos
except ImportError:
    from keras_tuner.protos import v3 as protos


def get_proto():
    return protos.keras_tuner_pb2


def get_service():
    return protos.service_pb2


def get_service_grpc():
    return protos.service_pb2_grpc



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup script."""

from setuptools import find_packages
from setuptools import setup

__version__ = "1.3.5"

setup(
    name="keras-tuner",
    description="A Hyperparameter Tuning Library for Keras",
    url="https://github.com/keras-team/keras-tuner",
    author="The KerasTuner authors",
    license="Apache License 2.0",
    version=__version__,
    install_requires=[
        "packaging",
        "requests",
        "kt-legacy",
    ],
    extras_require={
        "tensorflow": [
            "tensorflow>=2.0",
        ],
        "tensorflow-cpu": [
            "tensorflow-cpu>=2.0",
        ],
        "tests": [
            "black",
            "flake8",
            "isort",
            "ipython",
            "pandas",
            "portpicker",
            "pytest",
            "pytest-cov",
            "pytest-xdist",
            "namex",
            "scikit-learn",
            "scipy",
        ],
        "bayesian": [
            "scikit-learn",
            "scipy",
        ],
        "build": [
            "namex",
            "build",
        ],
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
    ],
    packages=find_packages(exclude=("*test*",)),
)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for HyperEfficientNet Model."""

import numpy as np
import pytest
import tensorflow as tf
from packaging.version import parse

from keras_tuner.applications import efficientnet
from keras_tuner.engine import hypermodel as hm_module
from keras_tuner.engine import hyperparameters as hp_module


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers and "
    "applications.efficientnet only exist in TF2.3+.",
)
@pytest.mark.parametrize("version", ["B0", "B1"])
def test_model_construction(version):
    hp = hp_module.HyperParameters()
    hp.Choice("version", [version])
    hypermodel = efficientnet.HyperEfficientNet(
        input_shape=(32, 32, 3), classes=10
    )
    model = hypermodel.build(hp)
    assert hp.values["version"] == version
    assert model.layers
    assert model.name == "EfficientNet"
    assert model.output_shape == (None, 10)
    model.train_on_batch(np.ones((1, 32, 32, 3)), np.ones((1, 10)))
    out = model.predict(np.ones((1, 32, 32, 3)))
    assert out.shape == (1, 10)


def test_tf_version_too_low_error():
    pp_module = efficientnet.preprocessing
    efficientnet.preprocessing = None

    with pytest.raises(ImportError, match="HyperEfficientNet requires"):
        efficientnet.HyperEfficientNet()

    efficientnet.preprocessing = pp_module


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers and "
    "applications.efficientnet only exist in TF2.3+.",
)
def test_hyperparameter_existence_and_defaults():
    hp = hp_module.HyperParameters()
    hypermodel = efficientnet.HyperEfficientNet(
        input_shape=(224, 224, 3), classes=10
    )
    hypermodel.build(hp)
    assert hp.get("version") == "B0"
    assert hp.get("top_dropout_rate") == 0.2
    assert hp.get("learning_rate") == 0.01
    assert hp.get("pooling") == "avg"


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers and "
    "applications.efficientnet only exist in TF2.3+.",
)
def test_hyperparameter_override():
    hp = hp_module.HyperParameters()
    hp.Choice("version", ["B1"])
    hp.Fixed("top_dropout_rate", 0.5)
    hypermodel = efficientnet.HyperEfficientNet(
        input_shape=(256, 256, 3), classes=10
    )
    hypermodel.build(hp)
    assert hp.get("version") == "B1"
    assert hp.get("top_dropout_rate") == 0.5


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers and "
    "applications.efficientnet only exist in TF2.3+.",
)
def test_input_tensor():
    hp = hp_module.HyperParameters()
    inputs = tf.keras.Input(shape=(256, 256, 3))
    hypermodel = efficientnet.HyperEfficientNet(input_tensor=inputs, classes=10)
    model = hypermodel.build(hp)
    assert model.inputs == [inputs]


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers and "
    "applications.efficientnet only exist in TF2.3+.",
)
def test_override_compiling_phase():
    class MyHyperEfficientNet(efficientnet.HyperEfficientNet):
        def _compile(self, model, hp):
            learning_rate = 0.1
            optimizer_name = hp.Choice(
                "optimizer", ["adam", "sgd"], default="adam"
            )
            if optimizer_name == "sgd":
                optimizer = tf.keras.optimizers.SGD(
                    momentum=0.1, learning_rate=learning_rate
                )
            elif optimizer_name == "adam":
                optimizer = tf.keras.optimizers.Adam(
                    learning_rate=learning_rate
                )
            model.compile(
                optimizer=optimizer,
                loss="categorical_crossentropy",
                metrics=["accuracy"],
            )

    hp = hp_module.HyperParameters()
    hypermodel = MyHyperEfficientNet(input_shape=(32, 32, 3), classes=5)
    hypermodel.build(hp)
    assert "learning_rate" not in hp.values
    assert hp.values["optimizer"] == "adam"


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers and "
    "applications.efficientnet only exist in TF2.3+.",
)
def test_augmentation_param_invalid_input():
    with pytest.raises(ValueError):
        efficientnet.HyperEfficientNet(
            input_shape=(32, 32, 3), classes=10, augmentation_model=0
        )


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers and "
    "applications.efficientnet only exist in TF2.3+.",
)
def test_augmentation_param_fixed_model():
    hp = hp_module.HyperParameters()
    aug_model = tf.keras.Sequential(name="aug")
    hypermodel = efficientnet.HyperEfficientNet(
        input_shape=(32, 32, 3), classes=10, augmentation_model=aug_model
    )
    model = hypermodel.build(hp)
    assert model.layers[1].name == "aug"


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers and "
    "applications.efficientnet only exist in TF2.3+.",
)
def test_augmentation_param_hyper_model():
    class HyperAug(hm_module.HyperModel):
        def build(self, hp):
            model = tf.keras.Sequential(name="aug")
            scaling_factor = hp.Choice("scaling_factor", [1])
            model.add(tf.keras.layers.Lambda(lambda x: x * scaling_factor))
            return model

    hp = hp_module.HyperParameters()
    aug_hm = HyperAug()
    hypermodel = efficientnet.HyperEfficientNet(
        input_shape=(32, 32, 3), classes=10, augmentation_model=aug_hm
    )
    model = hypermodel.build(hp)
    assert model.layers[1].name == "aug"
    assert hp.values["scaling_factor"] == 1


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers and "
    "applications.efficientnet only exist in TF2.3+.",
)
def test_pooling_is_max():
    hp = hp_module.HyperParameters()
    hp.values["pooling"] = "max"
    hypermodel = efficientnet.HyperEfficientNet(
        input_shape=(32, 32, 3), classes=10
    )
    hypermodel.build(hp)


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers and "
    "applications.efficientnet only exist in TF2.3+.",
)
def test_no_classes_raise_error():
    with pytest.raises(ValueError, match="classes"):
        efficientnet.HyperEfficientNet(input_shape=(32, 32, 3))


@pytest.mark.skipif(
    parse(tf.__version__) < parse("2.3.0"),
    reason="Preprocessing layers and "
    "applications.efficientnet only exist in TF2.3+.",
)
def test_no_input_shape_tensor_raise_error():
    with pytest.raises(ValueError, match="input_tensor"):
        efficientnet.HyperEfficientNet(classes=10)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from keras_tuner import protos
from keras_tuner.engine import hyperparameters as hp_module


def test_float():
    # Test with step arg
    linear = hp_module.Float(
        "linear", min_value=0.5, max_value=9.5, step=0.1, default=9.0
    )
    linear = hp_module.Float.from_config(linear.get_config())
    assert linear.default == 9.0
    assert 0.5 <= linear.random_sample() <= 9.5
    assert isinstance(linear.random_sample(), float)
    assert linear.random_sample(123) == linear.random_sample(123)

    # Test without step arg
    linear = hp_module.Float(
        "linear", min_value=0.5, max_value=6.5, default=2.0
    )
    linear = hp_module.Float.from_config(linear.get_config())
    assert linear.default == 2.0
    assert 0.5 <= linear.random_sample() < 6.5
    assert isinstance(linear.random_sample(), float)
    assert linear.random_sample(123) == linear.random_sample(123)

    # No default
    linear = hp_module.Float("linear", min_value=0.5, max_value=9.5, step=0.1)
    assert linear.default == 0.5


def test_float_linear_value_to_prob_no_step():
    rg = hp_module.Float("rg", min_value=1.0, max_value=11.0)
    assert abs(rg.value_to_prob(4.5) - 0.35) < 1e-4
    assert rg.prob_to_value(0.35) == 4.5


def test_float_log_with_step():
    rg = hp_module.Float(
        "rg", min_value=0.01, max_value=100, step=10, sampling="log"
    )
    for _ in range(10):
        assert rg.random_sample() in [0.01, 0.1, 1.0, 10.0, 100.0]
    assert abs(rg.value_to_prob(0.1) - 0.3) < 1e-4
    assert rg.prob_to_value(0.3) == 0.1


def test_float_reverse_log_with_step():
    rg = hp_module.Float(
        "rg", min_value=0.01, max_value=100, step=10, sampling="reverse_log"
    )
    for _ in range(10):
        # print(rg.random_sample())
        # assert rg.random_sample() in [0.01, 0.1, 1.0, 10.0, 100.0]
        # [0.09, 0.9, 9, 90]
        # [90, 9, 0.9, 0.09]
        sample = rg.random_sample()
        assert any(
            abs(sample - x) < 1e-4 for x in [0.01, 90.01, 99.01, 99.91, 100.0]
        )
    assert abs(rg.value_to_prob(99.91) - 0.3) < 1e-4
    assert abs(rg.prob_to_value(0.3) - 99.91) < 1e-4


def test_sampling_zero_length_intervals():
    f = hp_module.Float("f", 2, 2)
    rand_sample = f.random_sample()
    assert rand_sample == 2

    val = 2
    prob = f.value_to_prob(val)
    assert prob == 0.5


def test_log_sampling_random_state():
    f = hp_module.Float("f", 1e-3, 1e3, sampling="log")
    rand_sample = f.random_sample()
    assert rand_sample >= f.min_value
    assert rand_sample <= f.max_value

    val = 1e-3
    prob = f.value_to_prob(val)
    assert prob == 0
    new_val = f.prob_to_value(prob)
    assert np.isclose(val, new_val)

    val = 1
    prob = f.value_to_prob(val)
    assert prob == 0.5
    new_val = f.prob_to_value(prob)
    assert np.isclose(val, new_val)

    val = 1e3
    prob = f.value_to_prob(val)
    assert prob == 1
    new_val = f.prob_to_value(prob)
    assert np.isclose(val, new_val)


def test_reverse_log_sampling_random_state():
    f = hp_module.Float("f", 1e-3, 1e3, sampling="reverse_log")
    rand_sample = f.random_sample()
    assert rand_sample >= f.min_value
    assert rand_sample <= f.max_value

    val = 1e-3
    prob = f.value_to_prob(val)
    assert prob == 0
    new_val = f.prob_to_value(prob)
    assert np.isclose(val, new_val)

    val = 1
    prob = f.value_to_prob(val)
    assert prob > 0 and prob < 1
    new_val = f.prob_to_value(prob)
    assert np.isclose(val, new_val)


def test_float_sampling_arg():
    f = hp_module.Float("f", 1e-20, 1e10, sampling="log")
    f = hp_module.Float.from_config(f.get_config())
    assert f.sampling == "log"


def test_float_proto():
    hp = hp_module.Float("a", -10, 10, sampling="linear", default=3)
    proto = hp.to_proto()
    assert proto.name == "a"
    assert proto.min_value == -10.0
    assert proto.max_value == 10.0
    assert proto.sampling == protos.get_proto().Sampling.LINEAR
    assert proto.default == 3.0
    # Zero is the default, gets converted to `None` in `from_proto`.
    assert proto.step == 0.0

    new_hp = hp_module.Float.from_proto(proto)
    assert new_hp.get_config() == hp.get_config()


def test_float_values_property_with_step():
    assert list(hp_module.Float("float", 2, 8, 2).values) == [
        2.0,
        4.0,
        6.0,
        8.0,
    ]
    assert isinstance(list(hp_module.Float("float", 2, 8, 2).values)[0], float)
    assert list(
        hp_module.Float("float", 0.1, 100.0, 10, sampling="log").values
    ) == [
        0.1,
        1.0,
        10.0,
        100.0,
    ]


def test_float_values_property_without_step():
    assert len(list(hp_module.Float("float", 2, 4).values)) == 10
    assert len(list(hp_module.Float("float", 2, 20).values)) == 10
    assert (
        len(list(hp_module.Float("float", 2, 1024, sampling="log").values))
        == 10
    )



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

from keras_tuner.engine import hyperparameters as hp_module


def test_choice():
    choice = hp_module.Choice("choice", [1, 2, 3], default=2)
    choice = hp_module.Choice.from_config(choice.get_config())
    assert choice.default == 2
    assert choice.random_sample() in [1, 2, 3]
    assert choice.random_sample(123) == choice.random_sample(123)
    assert abs(choice.value_to_prob(1) - 1.0 / 6) < 1e-4
    # No default
    choice = hp_module.Choice("choice", [1, 2, 3])
    assert choice.default == 1
    with pytest.raises(ValueError, match="default value should be"):
        hp_module.Choice("choice", [1, 2, 3], default=4)


@pytest.mark.parametrize(
    "values,ordered_arg,ordered_val",
    [
        ([1, 2, 3], True, True),
        ([1, 2, 3], False, False),
        ([1, 2, 3], None, True),
        (["a", "b", "c"], False, False),
        (["a", "b", "c"], None, False),
    ],
)
def test_choice_ordered(values, ordered_arg, ordered_val):
    choice = hp_module.Choice("choice", values, ordered=ordered_arg)
    assert choice.ordered == ordered_val
    choice_new = hp_module.Choice(**choice.get_config())
    assert choice_new.ordered == ordered_val


def test_choice_ordered_invalid():
    with pytest.raises(ValueError, match="must be `False`"):
        hp_module.Choice("a", ["a", "b"], ordered=True)


def test_choice_types():
    values1 = ["a", "b", 0]
    with pytest.raises(TypeError, match="can contain only one"):
        hp_module.Choice("a", values1)
    values2 = [{"a": 1}, {"a": 2}]
    with pytest.raises(TypeError, match="can contain only `int`"):
        hp_module.Choice("a", values2)


def test_choice_value_not_provided_error():
    with pytest.raises(ValueError, match="`values` must be provided"):
        hp_module.Choice("a", [])


def test_choice_repr():
    assert repr(hp_module.Choice("a", [1, 2, 3])) == repr(
        hp_module.Choice("a", [1, 2, 3])
    )


def test_choice_none_as_default():
    hp = hp_module.Choice("a", [1, 2], default=None)
    assert hp.default == 1


def test_choice_default_not_none():
    hp = hp_module.Choice("a", [1, 2], default=2)
    assert hp.default == 2


def test_choice_proto():
    hp = hp_module.Choice("a", [2.3, 4.5, 6.3], ordered=True)
    proto = hp.to_proto()
    assert proto.name == "a"
    assert proto.ordered
    assert np.allclose([v.float_value for v in proto.values], [2.3, 4.5, 6.3])
    # Proto stores the implicit default.
    assert np.isclose(proto.default.float_value, 2.3)

    new_hp = hp_module.Choice.from_proto(proto)
    assert new_hp.name == "a"
    assert np.allclose(new_hp.values, hp.values)
    assert new_hp.ordered
    assert np.isclose(new_hp._default, 2.3)

    # Test int values.
    int_choice = hp_module.Choice("b", [1, 2, 3], ordered=False, default=2)
    new_int_choice = hp_module.Choice.from_proto(int_choice.to_proto())
    assert int_choice.get_config() == new_int_choice.get_config()

    # Test float values.
    float_choice = hp_module.Choice(
        "b", [0.5, 2.5, 4.0], ordered=False, default=2.5
    )
    new_float_choice = hp_module.Choice.from_proto(float_choice.to_proto())
    assert float_choice.get_config() == new_float_choice.get_config()


def test_prob_one_choice():
    hp = hp_module.Choice("a", [0, 1, 2])
    # Check that boundaries are valid.
    value = hp.prob_to_value(1)
    assert value == 2

    value = hp.prob_to_value(0)
    assert value == 0


def test_choice_values_property():
    assert list(hp_module.Choice("choice", [0, 1, 2]).values) == [0, 1, 2]



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the OracleServicer class."""

import os

import keras_tuner
from keras_tuner.distribute import oracle_chief
from keras_tuner.distribute import oracle_client
from keras_tuner.engine import metrics_tracking
from keras_tuner.test_utils import mock_distribute
from keras_tuner.tuners import randomsearch


def test_get_space(tmp_path):
    def _test_get_space():
        hps = keras_tuner.HyperParameters()
        hps.Int("a", 0, 10, default=3)
        oracle = randomsearch.RandomSearchOracle(
            objective=keras_tuner.Objective("score", "max"),
            max_trials=10,
            hyperparameters=hps,
        )
        oracle._set_project_dir(tmp_path, "untitled")
        tuner_id = os.environ["KERASTUNER_TUNER_ID"]
        if "chief" in tuner_id:
            oracle_chief.start_server(oracle)
        else:
            client = oracle_client.OracleClient(oracle)
            retrieved_hps = client.get_space()
            assert retrieved_hps.values == {"a": 3}
            assert len(retrieved_hps.space) == 1

    mock_distribute.mock_distribute(_test_get_space)


def test_update_space(tmp_path):
    def _test_update_space():
        oracle = randomsearch.RandomSearchOracle(
            objective=keras_tuner.Objective("score", "max"), max_trials=10
        )
        oracle._set_project_dir(tmp_path, "untitled")
        tuner_id = os.environ["KERASTUNER_TUNER_ID"]
        if "chief" in tuner_id:
            oracle_chief.start_server(oracle)
        else:
            client = oracle_client.OracleClient(oracle)

            hps = keras_tuner.HyperParameters()
            hps.Int("a", 0, 10, default=5)
            hps.Choice("b", [1, 2, 3])
            client.update_space(hps)

            retrieved_hps = client.get_space()
            assert len(retrieved_hps.space) == 2
            assert retrieved_hps.values["a"] == 5
            assert retrieved_hps.values["b"] == 1

    mock_distribute.mock_distribute(_test_update_space)


def test_create_trial(tmp_path):
    def _test_create_trial():
        hps = keras_tuner.HyperParameters()
        hps.Int("a", 0, 10, default=5)
        hps.Choice("b", [1, 2, 3])
        oracle = randomsearch.RandomSearchOracle(
            objective=keras_tuner.Objective("score", "max"),
            max_trials=10,
            hyperparameters=hps,
        )
        oracle._set_project_dir(tmp_path, "untitled")
        tuner_id = os.environ["KERASTUNER_TUNER_ID"]
        if "chief" in tuner_id:
            oracle_chief.start_server(oracle)
        else:
            client = oracle_client.OracleClient(oracle)
            trial = client.create_trial(tuner_id)
            assert trial.status == "RUNNING"
            a = trial.hyperparameters.get("a")
            assert a >= 0 and a <= 10
            b = trial.hyperparameters.get("b")
            assert b in {1, 2, 3}

    mock_distribute.mock_distribute(_test_create_trial)


def test_update_trial(tmp_path):
    def _test_update_trial():
        hps = keras_tuner.HyperParameters()
        hps.Int("a", 0, 10, default=5)
        oracle = randomsearch.RandomSearchOracle(
            objective=keras_tuner.Objective("score", "max"),
            max_trials=10,
            hyperparameters=hps,
        )
        oracle._set_project_dir(tmp_path, "untitled")
        tuner_id = os.environ["KERASTUNER_TUNER_ID"]
        if "chief" in tuner_id:
            oracle_chief.start_server(oracle)
        else:
            client = oracle_client.OracleClient(oracle)
            trial = client.create_trial(tuner_id)
            assert "score" not in trial.metrics.metrics
            trial_id = trial.trial_id
            client.update_trial(trial_id, {"score": 1}, step=2)
            updated_trial = client.get_trial(trial_id)
            assert updated_trial.metrics.get_history("score") == [
                metrics_tracking.MetricObservation([1], step=2)
            ]

    mock_distribute.mock_distribute(_test_update_trial)


def test_end_trial(tmp_path):
    def _test_end_trial():
        hps = keras_tuner.HyperParameters()
        hps.Int("a", 0, 10, default=5)
        oracle = randomsearch.RandomSearchOracle(
            objective=keras_tuner.Objective("score", "max"),
            max_trials=10,
            hyperparameters=hps,
        )
        oracle._set_project_dir(tmp_path, "untitled")
        tuner_id = os.environ["KERASTUNER_TUNER_ID"]
        if "chief" in tuner_id:
            oracle_chief.start_server(oracle)
        else:
            client = oracle_client.OracleClient(oracle)
            trial = client.create_trial(tuner_id)
            trial_id = trial.trial_id
            client.update_trial(trial_id, {"score": 1}, step=2)
            trial.status = "FAILED"
            client.end_trial(trial)
            updated_trial = client.get_trial(trial_id)
            assert updated_trial.status == "FAILED"

    mock_distribute.mock_distribute(_test_end_trial)


def test_get_best_trials(tmp_path):
    def _test_get_best_trials():
        hps = keras_tuner.HyperParameters()
        hps.Int("a", 0, 100, default=5)
        hps.Int("b", 0, 100, default=6)
        oracle = randomsearch.RandomSearchOracle(
            objective=keras_tuner.Objective("score", direction="max"),
            max_trials=10,
            hyperparameters=hps,
        )
        oracle._set_project_dir(tmp_path, "untitled")
        tuner_id = os.environ["KERASTUNER_TUNER_ID"]
        if "chief" in tuner_id:
            oracle_chief.start_server(oracle)
        else:
            client = oracle_client.OracleClient(oracle)
            trial_scores = {}
            for score in range(10):
                trial = client.create_trial(tuner_id)
                assert trial.status == "RUNNING"
                assert "a" in trial.hyperparameters.values
                assert "b" in trial.hyperparameters.values
                trial_id = trial.trial_id
                client.update_trial(trial_id, {"score": score})
                trial.status = "COMPLETED"
                client.end_trial(trial)
                trial_scores[trial_id] = score
            best_trials = client.get_best_trials(3)
            best_scores = [t.score for t in best_trials]
            assert best_scores == [9, 8, 7]
            # Check that trial_ids are correctly mapped to scores.
            for t in best_trials:
                assert trial_scores[t.trial_id] == t.score

    mock_distribute.mock_distribute(_test_get_best_trials, num_workers=1)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from keras_tuner import utils


def test_check_tf_version_error():
    utils.tf.__version__ = "1.15.0"

    with pytest.warns(ImportWarning) as record:
        utils.check_tf_version()
    assert len(record) == 1
    assert (
        "Tensorflow package version needs to be at least 2.0.0"
        in record[0].message.args[0]
    )


def test_to_list_with_tuple_return_list():
    result = utils.to_list((1, 2, 3))
    assert isinstance(result, list)
    assert result == [1, 2, 3]


def test_try_clear_without_ipython():
    is_notebook = utils.IS_NOTEBOOK
    utils.IS_NOTEBOOK = False
    utils.try_clear()
    utils.IS_NOTEBOOK = is_notebook


def test_create_directory_and_remove_existing(tmp_path):
    utils.create_directory(tmp_path, remove_existing=True)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import pytest

from keras_tuner.engine import conditions


def test_from_proto_error():
    with pytest.raises(ValueError, match="Unrecognized condition"):
        proto = mock.Mock()
        proto.WhichOneof.return_value = "unknown"
        conditions.Condition.from_proto(proto)


def test_unknown_condition_values_error():
    with pytest.raises(TypeError, match="Can contain only"):
        conditions.Parent("parent", [None])


def test_int_to_from_proto():
    condition = conditions.Parent("parent", [1, 2])
    condition_2 = conditions.Condition.from_proto(condition.to_proto())
    assert condition.values == condition_2.values


def test_float_to_from_proto():
    condition = conditions.Parent("parent", [1.0, 2.0])
    condition_2 = conditions.Condition.from_proto(condition.to_proto())
    assert condition.values == condition_2.values


def test_bool_to_from_proto():
    condition = conditions.Parent("parent", [True, False])
    condition_2 = conditions.Condition.from_proto(condition.to_proto())
    assert condition.values == condition_2.values



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_tuner.api_export import keras_tuner_export
from keras_tuner.engine import metrics_tracking


@keras_tuner_export("keras_tuner.Objective")
class Objective:
    """The objective for optimization during tuning.

    Args:
        name: String. The name of the objective.
        direction: String. The value should be "min" or "max" indicating
            whether the objective value should be minimized or maximized.
    """

    def __init__(self, name, direction):
        self.name = name
        self.direction = direction

    def has_value(self, logs):
        """Check if objective value exists in logs.

        Args:
            logs: A dictionary with the metric names as the keys and the metric
                values as the values, which is in the same format as the `logs`
                argument for `Callback.on_epoch_end()`.

        Returns:
            Boolean, whether we can compute objective value from the logs.
        """
        return self.name in logs

    def get_value(self, logs):
        """Get the objective value from the metrics logs.

        Args:
            logs: A dictionary with the metric names as the keys and the metric
                values as the values, which is in the same format as the `logs`
                argument for `Callback.on_epoch_end()`.

        Returns:
            The objective value.
        """
        return logs[self.name]

    def better_than(self, a, b):
        """Whether the first objective value is better than the second.

        Args:
            a: A float, an objective value to compare.
            b: A float, another objective value to compare.

        Returns:
            Boolean, whether the first objective value is better than the
            second.
        """
        return (a > b and self.direction == "max") or (
            a < b and self.direction == "min"
        )

    def __eq__(self, obj):
        return self.name == obj.name and self.direction == obj.direction

    def __str__(self):
        return f'Objective(name="{self.name}", direction="{self.direction}")'


class DefaultObjective(Objective):
    """Default objective to minimize if not provided by the user."""

    def __init__(self):
        super().__init__(name="default_objective", direction="min")


class MultiObjective(Objective):
    """A container for a list of objectives.

    Args:
        objectives: A list of `Objective`s.
    """

    def __init__(self, objectives):
        super().__init__(name="multi_objective", direction="min")
        self.objectives = objectives
        self.name_to_direction = {
            objective.name: objective.direction for objective in self.objectives
        }

    def has_value(self, logs):
        return all(key in logs for key in self.name_to_direction)

    def get_value(self, logs):
        obj_value = 0
        for metric_name, metric_value in logs.items():
            if metric_name not in self.name_to_direction:
                continue
            if self.name_to_direction[metric_name] == "min":
                obj_value += metric_value
            else:
                obj_value -= metric_value
        return obj_value

    def __eq__(self, obj):
        if self.name_to_direction.keys() != obj.name_to_direction.keys():
            return False
        return sorted(self.objectives, key=lambda x: x.name) == sorted(
            obj.objectives, key=lambda x: x.name
        )

    def __str__(self):
        return (
            "Multi"
            + super().__str__()
            + f": [{', '.join(map(lambda x: str(x), self.objectives))}]"
        )


def create_objective(objective):
    if objective is None:
        return DefaultObjective()
    if isinstance(objective, list):
        return MultiObjective([create_objective(obj) for obj in objective])
    if isinstance(objective, Objective):
        return objective
    if not isinstance(objective, str):
        raise TypeError(
            "`objective` not understood, expected str or "
            f"`Objective` object, found: {objective}"
        )

    direction = metrics_tracking.infer_metric_direction(objective)
    if direction is None:
        error_msg = (
            'Could not infer optimization direction ("min" or "max") '
            'for unknown metric "{obj}". Please specify the objective  as'
            "a `keras_tuner.Objective`, for example `keras_tuner.Objective("
            '"{obj}", direction="min")`.'
        )
        error_msg = error_msg.format(obj=objective)
        raise ValueError(error_msg)
    return Objective(name=objective, direction=direction)
